; Calarm.asm ( ASSEMBLER SOURCE CODE MODULE )
; Installs a TSR handler for Int 4ah and sets a hardware alarm
;       using int 1Ah. The handler prints a message and rings the bell
;       This module is called by a C-program via an external function
;       setup(int time,...)  where time gives the number of seconds from now
;       to set the alarm. If time is 0 the alarm is set at a given time of day
;       stored in the public variables ahours, aminutes, aseconds. These
;       are set from the C-module if the /t option is invoked.
;       Otherwise,  hours now plus the time offset
;       from now until the alarm, must be stored in an 8 bit register, hence
;       cannot exceed 256. Thus the argument to setup may not safely exceed
;       256 - 24 = 232 hours

; Interrupt to take over: timer alarm --
intr equ 4AH


_TEXT SEGMENT PUBLIC 'CODE'
ASSUME cs:_TEXT,ds:_TEXT
org 100h
begin:            ; Program entry point
jmp near ptr crt  ; Jump around to beginning of C-run time

PUBLIC _time_s,_time_m,_time_h,_hours,_minutes,_seconds,_ahours,_aminutes,_aseconds,_too_long,_bell,_banner
msg     db 80 dup (' ')   ; default message is 80 chars of blanks
; later: move to C-pgm
leng    dw  80
err_msg_1    db 'Cannot install: Busy, or clock error',0DH,0AH,'$'
err_msg_2    db 'Cannot uninstall. No, or wrong 01H or 4AH vectors present.',0DH,0AH,'$'
motto db '                         CHICKEN HAVEN SOFTWARE',0DH,0AH,'$' 
install db '        Alarm 1.2: By Terry McConnell (1996) No rights reserved.',0DH,0AH,'$'
old4avoff dw 00H ; address of old int 4ah handler: for uninstall
old4avseg dw 00H
old01voff dw 00H ; address of old int 01H handler: for uninstall
old01vseg dw 00H
base  db 10      ; for a div in convert routine
parlen db 16     ; for a div
_time_s  db 0    ; alarm time offset, seconds
_time_m  db 0    ;  ... minutes
_time_h  db 0    ;  ... hours
secs_per_min  dw 60 ; for a div
sec_carry db 0   ; storage for a carry digit
min_carry db 0
_too_long dw 0   ; flag to signal that 2 or more midnight passages will occur
_bell db 1 ; flag to signal whether the bell should be rung
_banner dw 1; flag to signal that the start-up info should be printed

; UNITIALIZED DATA
PSP_SEG dw ?           ; segment address of our PSP
time_lo dw ?
	; low word of number of seconds till alarm
time_hi dw ?    ;  hi word of number of seconds till alarm
_hours    db ?        ; time of day now (military style )
_minutes db  ?
_seconds  db ?
_ahours  db ?         ; time of alarm
_aminutes db ?
_aseconds db ?
uninst_seg_hi db ?    ; Used to store the seg:offset address of the uninst
uninst_seg_lo db ?    ; label in the tsr, which is the entry point for our
uninst_off_hi db ?    ; int 01 handler. Since segmented addresses are not
uninst_off_lo db ?    ; unique we need this exactly the way it was stored

PUBLIC _setup,_tsr,_uninstall
;
; BEGINNING  OF RESIDENT ROUTINE
;       handle int4ah, handle int 01h in order to send data back to
;       non-resident program for uninstall purposes.
;
_message: sti      ; enable interrupts.
	push ds    
	push ax    ; Save all registers on the stack
	push si    ; These are all the registers that will be used
	push cx
        mov ax,_TEXT  ; Load ds so we can access our data
	mov ds,ax
;
; CHAIN TO OLD 4AH HANDLER: Generally good form to do this.
;
	pushf                     ; The first 3 pushes get us back here when
	push seg back             ; the old handler iret's.
	push offset back   
	push old4avseg            ; I got this trick to make a far jump 
	push old4avoff            ; using retf off the net. (Clever!)
	retf                      ; Going to old handler now. Bye!
;
;
back:
        mov ah,0Eh      ; back from old handler now, hopefully.
        mov al,_bell    ; check to see whether we should ring the bell
        cmp al, 0
        jz end_bell        
	mov al,07H     ; Ring the bell using
	int 10H        ; ROM BIOS video service
end_bell:
	mov cx,leng
	lea si,msg
pr_char:    lodsb                ; print out message
	int 10H                  ; byte by
	loop pr_char             ; byte.
;
; PREPARE TO RETURN FROM INTERRUPT  4aH
;
	pop cx
	pop si        ; restore registers used
	pop ax
	pop ds
	iret
;
; INT 01H HANDLER
;
; The routine below is invoked from alarm (new incarnation) via int01
; Its job is to return the address of the original int 01H handler
; in di:dx and the original int4a handler in bx:cx
; so that these can be restored in the uninstall routine of alarm.
; Also sends back cs:_message in bp:ax and cs:_uninst in bp:es
; so that these can be compared with the current interrupt vectors
; for int 01H and int 4aH. If they don't agree, then some other routine
; has stolen these interrupts, and it's not safe to uninstall.

uninst: sti
	push ds         ; save registers used 
	push ax         ; in case the interrupt was not generated
			; by our program and we need to bail out.
        mov ax,_TEXT    ; load ds so we can access our data
	mov ds,ax
;
; CHAIN TO OLD 01H HANDLER
;
	pushf                     ; The first 3 pushes get us back here when
	push seg back_again       ; the old handler iret's.
	push offset back_again   
	push old01vseg            
	push old01voff           
	retf                      ; Going to old handler now. Bye!

back_again:             
;
;  Now we are going to send back information in the registers, but we
;  wouldn't want to do this if some other program had taken over
;  int 01H -- it wouldn't be expecting what we are going to send.
;  So check to make sure that the current 01H interrupt handler points
;  to uninst.  

	mov ah,0000:0007          ; seg of 01H handler
	cmp ah,uninst_seg_hi      ; compare with our saved uninst address
	jne no_can_do             ; and abort the uninstall if they are
	mov al,0000:0006          ; are not the same.
	cmp al, uninst_seg_lo
	jne no_can_do
	mov ah,0000:0005          ; offset of 01H handler
	cmp ah, uninst_off_hi
	jne no_can_do
	mov al,0000:0004
	cmp al, uninst_off_lo
	jne no_can_do
	jmp cont                  ; ok to continue with uninstall

no_can_do:
	pop ax      ; restore in case we are returning to some other
	pop ds      ; program (e.g., debug.)
	iret

cont:   pop ax               ; throw away: we're returning to our own program
	mov ax,PSP_SEG       ; load registers with all our info
	mov si,ax
	mov ax,old01voff
	mov dx,ax
	mov ax,old4avseg
	mov bx,ax
	mov ax,old4avoff
	mov cx,ax
	mov ax,old01vseg
	mov di,ax
	mov ax,_TEXT
	mov bp,ax
	mov ax,offset uninst
	mov es,ax
	mov ax,offset _message
	pop ds
	iret

endtsr:            ; end of resident routine 
;                   
;                   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; TRANSIENT ("SETUP") PART OF PROGRAM
; This gets called from a C-program
;
_setup  proc    ; beginning of "transient" routine
	push bp
	mov bp,sp
        mov ax,_TEXT   ; point ds at data segment
	mov ds,ax
;
;  UNLOAD PARAMETERS FROM THE STACK
;
	mov ax,[bp+4]       ; get parameter from C program (32 bit int)
	mov time_lo,ax      ; first off is the time offset (2 words)
	mov ax,[bp+6]       ; in seconds
	mov time_hi,ax
	; Next on the stack is a pointer to a string. Load source from it
	mov ax,[bp+8]
	mov si,ax
	mov di, offset msg   ; copy destination is msg
	mov ax,[bp+10]     ; length of string to copy comes next
	mov cx,ax          ; load counter with it
copyloop:
	movsb              ; copy the user-supplied message
	loop copyloop

;
;  SAVE SEGMENT ADDRESS OF OUR PSP (for uninstall)
;
	mov ah,51H    ; Dos function: Get PSP address
	int 21H
	mov PSP_SEG,bx    ;segment address of current PSP
;
;  SAVE ADDRESS OF OLD INTERRUPT 4AH HANDLER (for uninstall)
;

	mov al,intr  ; save address of old 4a handler
	mov ah,35H   ; using dos function 35h
	int 21H
	mov ax,es
	mov old4avseg,ax   ; save segment register of handler
	mov ax,bx
	mov old4avoff,ax   ; save offset of handler
	mov al,01H  ; save address of old 01H handler
	mov ah,35H   ; using dos function 35h
	int 21H
	mov ax,es
	mov old01vseg,ax   ; save segment register of handler
	mov ax,bx
	mov old01voff,ax   ; save offset of handler
	push ds
;
; INSTALL NEW HANDLER for 4aH (The resident part of this program: message)
;
	mov ax,cs        ; point ds:dx to new handler
	mov ds,ax        ; in our code segment
	mov dx, offset _message
	mov al,intr   ; put number of interrupt to hook in al
	mov ah,25H    ; point to handler in ds:dx
	int 21H
;
; INSTALL NEW HANDLER FOR 01H ( the resident part of this program: uninst)
       mov ax,cs       ; we need to do this again because 21h
       mov ds,ax       ; has trashed ds ?
       mov dx, offset uninst
       mov al, 01H
       mov ah,25H
       int 21H

	pop ds      ; points at our _data again
	mov al,0000:0007
	mov uninst_seg_hi,al
	mov al,0000:0006
	mov uninst_seg_lo,al
	mov al, 0000:0005
	mov uninst_off_hi,al
	mov al, 0000:0004
	mov uninst_off_lo,al
;
;   TURN OFF ALARM IF SET PREVIOUSLY
;
	mov ah,07H   ; turn off alarm if set previously
	int 1AH
;
;   COMPUTE TIME OF DAY ALARM WILL GO OFF  (It's given as seconds from now)
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; GET TIME OF DAY using BIOS 1AH service 02H
; Store in public variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
timenow:  mov ah, 02H  ; Bios 1AH service number
          int 1AH      ; BIOS interrupt
                       ; hours (packed bcd) are left in ch
          mov al,ch
          call bcdbin ; convert to binary
          mov _hours, al
          mov al, cl   ; minutes (bcd) in cl
          call bcdbin ; convert to binary
          mov _minutes,al
          mov al,dh    ; seconds (bcd) in dh
          call bcdbin
          mov _seconds, al
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; convert time (word, in seconds) to hours, minutes, seconds
alarmtime:
	mov ax, time_lo
	mov dx, time_hi    ;prepare to divide
        cmp ax,0
        jnz  not_both ;  if both ax and dx are zero skip finding ahours etc
        cmp dx,0
        jnz not_both;
                           ;skip over  setting a_hours, etc if time_lo
                           ;and time_hi are both zero. In that case they
        jmp set_alarm      ;are set from main.

not_both:
	div secs_per_min; divide by 60 
	mov _time_s,dl    ; remainder of division (< 60) is in dx
	cmp ax,0         ; see if quotient is zero
	jz over         ; done, if so
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; repeat the whole thing for minutes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	sub dx,dx       ;prepare to divide again
	div secs_per_min ; divide by 60
	mov _time_m,dl
	cmp ax,0         ; see if quotient is zero
	jz over         ; done if so
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Hours left over are now in al.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
over:      mov _time_h,al    
; ALARM OFFSET NOW IN time_h,time_m,time_s

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  Now add the offset to the time-now, with appropriate carrys
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        mov al,_seconds    ; time_now seconds to al
	add al,_time_s     ; add offset
	cmp al,60         ; see if result is more than 60
	jb no_adjust_s     ; don't need to adjust minutes and secs
	add sec_carry,1    ; else, save to add an extra minute to alarm time
	sub al,60          ; and take away 60 seconds
no_adjust_s:
	mov _aseconds,al    ; save alarm time, seconds
	mov al,_minutes      ; time_now minutes (adjusted) to al
	add al,_time_m       ; add offset
	add al,sec_carry     ; add extra minute if needed
	cmp al,60           ; see if result is 60 or more
	jb no_adjust_m      ; don't need to adjust minutes and hours
	add min_carry,1     ; else, save to add an extra hour to alarm time
	sub al,60           ; and take away 60 minutes
no_adjust_m:
	mov _aminutes,al      ; save alarm time, minutes
	mov al,_hours
	add al,_time_h
	add al,min_carry        ; add the extra hour if necessary
	cmp al,23               ; see if we go past midnight
	jna no_adjust_h          ; we don't, so don't change anything
	sub al,24      ; we do, so subtract 24 hours from alarm time
no_adjust_h:
	mov _ahours,al       ; save alarm time, hours
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Check whether ahours is still > 23 and warn
; the user, if so, that the alarm will go off a day earlier than she thinks
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	cmp al,23
	jbe no_warn
comp:
	sub al,24     ; reduce to time on 24 hour clock (i.e mod 24 )
	cmp al,23
	ja comp
	mov _ahours,al
	mov ax,1
	mov _too_long,ax    ; set flag to signal condition to C-program

no_warn:

set_alarm:
        mov al,_ahours
	call convert       ;  The service below assumes BCD values in regs
	mov ch,al     ; set alarm to go off at ...
	mov al,_aminutes
	call convert
	mov cl,al
	mov al,_aseconds
	call convert
	mov dh,al
	mov ah,06H     ; set alarm service  of BIOS int 1AH                  
	int 1AH
;
; ALARM IS NOW SET
	jc barf     ; Service 1AH sets the carry bit when there is a problem
        cmp _banner,1
        jne gohome
	mov dx, offset motto
	mov ah,09H
	int 21H           ;Print our motto
	mov dx, offset install
	mov ah,09H
	int 21H           ; Print our copyright
gohome: 
	pop bp            ; restore bp
	ret

barf:       
	mov dx,offset err_msg_1
	mov ah,09H    ; display string DOS service
	int 21H       ; DOS
	pop bp         ; restore bp
	mov al,01H     ; return value signals type 1 error
	mov ah,4CH    ; return to DOS without installing TSR.
	int 21H         ; exit to DOS

_setup   endp

; tsr:
; sets up a terminate and stay resident program to handle the alarm

_tsr proc near

; Free the program's environment block. The tsr does not use it

        push es
        mov ax, PSP_SEG
        mov es,ax        ; store psp seg address in es
        mov ax, es:002CH  ;seg address of env block at offset 2CH in psp
        mov es,ax
        mov ah,49H        ; free allocated memory (environment block)
        int 21H
        pop es

	mov ax,offset endtsr  ; compute number of paragraphs to keep in memory
	div parlen     ; divide by 16 = paragraph length in bytes
        xor ah,ah      ; clear remainder
        add ax,12h     ; convert to offset from PSP
        mov dx,ax  ; 31h wants number of paragraphs to keep in dx
        xor al,al     ; make the return value 0
        mov ah,31h    ; keep program return to DOS
	int 21H
_tsr endp

;
;  UNINSTALL ROUTINE
;

;  The installed tsr contains an int 01H handler whose purpose is to
;  return some stored data in the registers. The data returned are
;  si: segment address of the resident program's PSP
;  di:dx address of the resident program's old int 01H handler
;  bx:cx address of the resident program's old int 4aH handler
;  bp:ax address of resident programs own int01 handler
;  bp:es address of resident programs own int4a handler
;  This routine refuses to uninstall the tsr if either of the
;  handler addresses returned on the stack 
;  is different from the current interrupt vector
;  handler assignments-- that would mean some other handler had been
;  installed in the meantime. Assuming that one chains back to ours,
;  we had better not remove it. (Hint from Marc Johnson.)

_uninstall proc near
        mov ax,_TEXT   ; point ds at data segment
	mov ds,ax
	mov al,intr  ; save address of current 4ah handler
	mov ah,35H   ; using dos function 35h
	int 21H
	mov ax,es
	mov old4avseg,ax   ; save segment register of handler
	mov ax,bx
	mov old4avoff,ax   ; save offset of handler
	mov al,01H  ; save address of current 01h handler
	mov ah,35H   ; using dos function 35h
	int 21H
	mov ax,es
	mov old01vseg,ax   ; save segment register of handler
	mov ax,bx
	mov old01voff,ax   ; save offset of handler
	; stuff below trashes ds, so save it
	;
	push ds
	push si
	push di
	push es
	push bp

	;  OK, now we can uninstall our handler
; We must put everything back the way we found it: the old int4ah and
; int01h vectors must be restored, environment memory must be freed,
; and the program memory must be freed.
	int 01H
	cmp ax, [old4avoff]
        jne err2
	mov ax,es
	cmp ax, [old01voff]
        jne err2
	mov ax,bp
	cmp ax, [old4avseg]
        jne err2
	cmp ax, [old01vseg]
        jne err2
	mov ax,di       ; di:dx has address of old handler
	mov ds,ax
	mov al,01H
	mov ah,25H
	int 21H         ; restore old int 01H handler
	mov ax,bx
	mov ds,ax
	mov ax,cx
	mov dx,ax
	mov al,4aH
	mov ah,25H
	int 21H          ; restore old int 4a handler
	mov ax,si
	mov es,ax     ; seg address of psp
	mov ah,49H
	int 21H       ; free allocated memory
	pop bp
	pop es
	pop di
	pop si
	pop ds
	ret           ; to main ( where a success message is printed.)

err2:    mov dx, offset err_msg_2          ; print an error message and exit
	mov ah,09H
	int 21H
	mov al,01H    ; return value is error 1
	mov ah,4CH
	int 21H       ; return to DOS
_uninstall endp

; convert binary in al to BCD there
convert proc near
conv:  push cx
       xor ah,ah    
       div base
       mov ch,al    ; save first digit
       mov cl,4     ; number of binary places to shift
       sal ch,cl    ; shift 
       mov al,ah
       add al,ch
       pop cx
       ret
convert endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; bcdbin: convert packed bcd in al to binary there -- trashes ah
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
bcdbin proc near
        push dx     ; save registers used
        push cx
        mov dl, al  ; copy al
        and dl, 0FH ; low digit in dl
        and al, 00F0H
        mov cl,4    ; number of places to shift
        shr al,cl   ; hi digit now in al
        mul base    ; al now has (hi-digit * base )
        add al,dl   ; hi-digit*base + lo-digit
        pop cx
        pop dx      ; restore registers
        ret
bcdbin endp

crt:       ; Borland C run-time stuff goes below
_TEXT ends

sseg SEGMENT STACK 'DATA'
        dw 80 dup (?)
sseg ends

end begin
