/* alarm.c: calls an assembly module to set an alarm using   
	ROM Bios calls and TSR techniques   */

        /* VERSION 1.2 By Terry McConnell, June 1996 */

/* Needs to be linked with calarm11.obj and atosecs.obj. The sources for
   these are called calarm11.asm and atosecs.c  */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>




#define DEFAULT_TIME 13    /* A short 13 second default for debugging */ 
#define INFO  "\nAlarm: Set an alarm to go off after a specified time.\n\n\
usage:  alarm [/help /u /t /q /s /m <message> ] <time offset> \n\
   where /h[elp] or /? prints this message,\n\
         /u uinstalls the tsr that rings the alarm, \n\
         /q signals that the bell should not be rung, \n\
         /s signals that the start-up info should not be printed,\n\
         /t signals that the time given is a time of day, not an offset, and \n\
         /m customizes the message printed at alarm time.\n\n\
Options q,s,t and m can be combined, e.g /qsm\n\
The time offset should not go past midnight twice or more.\n\
A time offset should not exceed 231 hours.\n\
It can be given as a number of seconds or in the \n\
form xx:xx:xx.x  Fractions of a second are rounded."   
		
#define MAX_ARG_LEN 80             /* Max length of combined options */
#define DEFAULT_MSG "Smile :-) The alarm has rung!\n"
#define MSG_LEN 80
#define USAGE "Usage: alarm [ /help /u /t /q /s /m <string> ] <time>\n"
#define MAX_SECONDS 831600  /* This is 255 - 24 hours = 231 hours */
#define ERR3 "Cannot track midnight passages. Alarm will occur day(s) early."

int find_opt(char *option_string);
int setcmp(const char *, const char *, char);
extern setup(unsigned long int time,char *msg,int length);
extern short unsigned  int time_s,time_m,time_h,hours,minutes,seconds,ahours,aminutes,aseconds;
extern void tsr(void);
extern void uninstall(void);
extern unsigned long int atosecs(char *);
extern short int bell;
extern int too_long;

int offset = 1; /* flag to indicate meaning of last command line arg */
extern short int banner; /* Yes, we do print start-up info */

int
main(int argc, char **argv)
{
	unsigned long int time = DEFAULT_TIME;
	int c=-1;                /* numeric option code */
	char *info = INFO;
	char argstr[MAX_ARG_LEN] = "/";
	char *msg = DEFAULT_MSG;
	int length = strlen(DEFAULT_MSG);



	if(--argc > 0){  /* Anything on command line ? */
		if((*++argv)[0]=='/'){      /* already the end of options */
			if((*argv)[1]=='/'){
				++argv;
				c=-1;  /* no options */
			}
			else strcat(argstr,++*argv); /* start building argstr*/
			argc--;
			while((argc>0)&&((*++argv)[0]=='/')&&((*argv)[1]!='/')){
/* combine all options into a single string */
				argc--;
				strcat(argstr,++*argv);
			}
                        c=find_opt(argstr); /* look up code, set flags */
		}
		switch(c){

			case -1: break;  /* default, e.g no options */

			case 0: printf("%s\n",info);
				return 0;

			case 1: uninstall();
				printf("Alarm deactivated. Memory freed.\n");
				return 0;
                        case 2: /* A custom message is provided */
                                if(argc < 1) {
					fprintf(stderr,USAGE);
					return 1;
				}
				msg = *argv;
				++argv;
				--argc;
				length = strlen(msg);
				if(length > MSG_LEN){
					fprintf(stderr,"Message can only be %d characters long\n",MSG_LEN);
					return 1;
				}
				break;
                        case 3: /* no custom msg. flags have been set */
                                break;
			case -2: ;  /* barf (don't use in your code) */
			default:
				fprintf(stderr,"%s: illegal option\n",argstr);
				return 1;
			}
	}
	if(argc > 0)time = atosecs(*argv);
        if(time == 0){
		  fprintf(stderr,USAGE);
		  return 1;
	}
	if(time > MAX_SECONDS) {  /* Too big */
	fprintf(stderr,"alarm: time cannot exceed %lu secs.\n",MAX_SECONDS);
		return 1;
	}
        if(offset)
                setup(time,msg,length);
        else { /* need to set the time of day of alarm, ahours, aminutes, etc
                  from here rather than from setup */
             if(time > (unsigned long)86400){  /* seconds in a day */
                fprintf(stderr,"alarm: time of day out of range.\n");
                return 1;
             }
             ahours = time/3600;     /* number of even hours */
             time %= 3600;           /* number of seconds remaining */
             aminutes = time/60;
             aseconds = time % 60;
            

             /* A call to setup with first parameter = 0 informs it that
                it should skip the computation of ahours, etc. */
             setup((unsigned long)0,msg,length);
       }

        if(banner){
         /* Print out information on the alarm that has been set */
         /* the % 256 converts an unsigned word to an unsigned byte */

         printf("Alarm set at: %hd:%02hd:%02hd\n",hours %  256,minutes % 256,seconds % 256);
         printf("To go off: %hd:%02hd:%02hd\n",ahours % 256,aminutes % 256,aseconds % 256);    
         if(offset)
         printf("Which is %hd:%02hd:%02hd from now.\n",time_h % 256,time_m % 256,time_s % 256);
       }

         if(too_long != 0) /* 2 or more midnight passages occur */
		fprintf(stderr,"alarm: %s\n",ERR3);

	/* Install a tsr to trap interrupt 4ah which is generated by the
		BIOS at alarm time */

	tsr(); /* return to DOS via assembly module */

	/* You should never get to here. */
	return 0;  /* just to avoid compiler warning :-) */
}
  

/*  find_opt: return a unique small integer for each possible option string */
/*  There should be a case in the switch statement in main to handle each. */
/*  -2 reserved for arg not found -1 reserved for no options */

int find_opt(char *word)
{
	if(strcmp(word,"/help")==0)return 0;
	if(strcmp(word,"/?")==0)return 0;
	if(strcmp(word,"/h")==0)return 0;
	if(strcmp(word,"/u")==0)return 1;
	if(strcmp(word,"/m")==0)return 2;
        if(strcmp(word,"/s")==0){
                banner = 0;
                return 3;
        }
        if(setcmp(word,"/sm",'\1')){
           banner = 0;
           return 2;
        }
        if(strcmp(word,"/t") ==0){
          offset = 0;
          return 3;
        }
        if(setcmp(word,"/ts",'\1')){
                banner = 0;
                offset = 0;
                return 3;
        }
        if(setcmp(word,"/tm",'\1')){
           offset = 0;
           return 2;
        }
        if(setcmp(word,"/tsm",'\1')){
                offset = 0;
                banner = 0;
                return 2;
        }
        if(strcmp(word,"/q")==0){
           bell = 0;
           return 3;
        }
        if(setcmp(word,"/qs",'\1')){
                bell = 0;
                banner = 0;
                return 3;
        }
        if(setcmp(word,"/qt",'\1')){
           offset = 0;
           bell = 0;
           return 3;
        }
        if(setcmp(word,"/qst",'\1')){
           bell = 0;
           banner =0;
           offset =0;
           return 3;
       }
        if(setcmp(word,"/qtm",'\1')){
           offset = 0;
           bell = 0;
           return 2;
        }
        if(setcmp(word,"/qstm",'\1')){
           offset = 0;
           bell = 0;
           banner = 0;
           return 2;
       }
        if(setcmp(word,"/qm",'\1')){
           bell = 0;
           return 2;
        }
        if(setcmp(word,"/qsm",'\1')){
           bell = 0;
           banner = 0;
           return 2;
        }
/* args not found */
	return -2;
}


/* setcmp: string function that returns 1 if s and t are equal as sets,
           i.e., contain the same characters, not necessarily in the
           same order. Returns 0 if not, or if the answer cannot be
           determined. The invld character is any character not present
           in either s or t.  */

int setcmp(const char *s, const char *t, char invld)
{
        int i;
        char c;
        char *ptr;
        char buffer[MAX_ARG_LEN];

        if(s == NULL && t == NULL) return 1;       /* trivial case */
        if( t == NULL ) return 0; /* since here s must not be null */
        if(strchr(t,invld)!=NULL) {
                fprintf(stderr,"setcmp: error: character %c occurs in string %s\n",
                        invld,t);
               return 0;
        }
        if(strlen(t) != strlen(s)) return 0;

        /* From now on, both s and t are non-empty and of the same length */
        strcpy(buffer,t);
        for(i=0;i<strlen(s);i++){  /* loop over chars in s */
                c = *(s+i);  /* get next char of s */
                if((ptr=strchr(buffer,c))==NULL) return 0; /* No c in t */
                *ptr = invld;    /* else replace first occurence of c with
                                     invld so it won't be found again */
        }
        return 1; /* OK if we got to here */
}


